import unittest
from dtflw.logger import DefaultLogger
from dtflw.flow_context import FlowContext
from dtflw.input_table import InputTable
import tests.utils as utils


class InputTableTestCase(unittest.TestCase):

    def test_validate_succeeds(self):

        input_name = "foo"
        input_file_name = f"{input_name}.parquet"
        input_file_path = f"mock:/{input_file_name}"

        dbutils_mock = utils.mock_dbutils(
            fs_ls_raises_error=False,
            ls_return=[(input_file_name, input_file_path)]
        )
        storage = utils.StorageMock("", None, dbutils_mock)

        i = InputTable(
            name=input_name,
            abs_file_path=input_file_path,
            ctx=FlowContext(
                storage=storage,
                spark=None,
                dbutils=dbutils_mock,
                logger=DefaultLogger()
            )
        )

        i.validate()

    def test_validate_fails_file_path_not_given(self):

        i = InputTable(name="foo", abs_file_path=None, ctx=None)

        with self.assertRaises(Exception):
            i.validate()

    def test_validate_fails_file_does_not_exist(self):

        dbutils_mock = utils.mock_dbutils(fs_ls_raises_error=True)
        storage = utils.StorageMock("", None, dbutils_mock)

        i = InputTable(
            name="foo",
            abs_file_path=storage.get_abs_path("foo.parquet"),
            ctx=FlowContext(
                storage=storage,
                spark=None,
                dbutils=dbutils_mock,
                logger=DefaultLogger()
            )
        )

        with self.assertRaises(Exception):
            i.validate()

    def test_validate_is_optional_does_not_fail_file_does_not_exist(self):

        dbutils_mock = utils.mock_dbutils(fs_ls_raises_error=True)
        storage = utils.StorageMock("", None, dbutils_mock)

        i = InputTable(
            name="foo",
            abs_file_path=storage.get_abs_path("foo.parquet"),
            ctx=FlowContext(
                storage=storage,
                spark=None,
                dbutils=dbutils_mock,
                logger=DefaultLogger()
            ),
            is_optional=True
        )
        i.validate()



import unittest
from dtflw import init_args, init_inputs, init_outputs
from unittest.mock import patch
import tests.utils as utils
from ddt import ddt, data, unpack
import dtflw.com as com
import dtflw.arguments as a


@ddt
class ArgumentTestCase(unittest.TestCase):

    @data(
        (a.Argument, True),
        (a.Input, False),
        (a.Output, False)
    )
    @unpack
    @patch("dtflw.databricks.get_this_notebook_abs_path")
    @patch("dtflw.databricks.get_spark_session")
    @patch("dtflw.databricks.get_dbutils")
    def test_create_with_names_no_values(self, argument_type, has_value, get_dbutils_mock, get_spark_session_mock, get_this_notebook_abs_path_mock):
        get_dbutils_mock.return_value = utils.DButilsMock()
        get_spark_session_mock.return_value = utils.SparkSessionMock()

        notebook_abs_path = "/Repos/user@a.b/project/nb"
        get_this_notebook_abs_path_mock.return_value = notebook_abs_path

        actual = argument_type.create("foo", "bar")

        self.assertEqual(2, len(actual))

        self.assertEqual(actual["foo"].value, "")
        self.assertEqual(actual["foo"].has_value, has_value)

        self.assertEqual(actual["bar"].value, "")
        self.assertEqual(actual["bar"].has_value, has_value)

    @data(a.Argument, a.Input, a.Output)
    @patch("dtflw.databricks.get_this_notebook_abs_path")
    @patch("dtflw.databricks.get_dbutils")
    def test_create_with_names_and_default_values(self, argument_type, get_dbutils_mock, get_this_notebook_abs_path_mock):
        get_dbutils_mock.return_value = utils.DButilsMock()

        notebook_abs_path = "/Repos/user@a.b/project/nb"
        get_this_notebook_abs_path_mock.return_value = notebook_abs_path

        actual = argument_type.create({"foo": "a", "bar": "b"})

        self.assertEqual(2, len(actual))

        self.assertEqual(actual["foo"].value, "a")
        self.assertTrue(actual["foo"].has_value)

        self.assertEqual(actual["bar"].value, "b")
        self.assertTrue(actual["bar"].has_value)

    @patch("dtflw.databricks.get_this_notebook_abs_path")
    @patch("dtflw.databricks.get_spark_session")
    @patch("dtflw.databricks.get_dbutils")
    def test_create_with_names_and_shared_values(self, get_dbutils_mock, get_spark_session_mock, get_this_notebook_abs_path_mock):
        get_dbutils_mock.return_value = utils.DButilsMock()
        get_spark_session_mock.return_value = utils.SparkSessionMock()

        notebook_abs_path = "/Repos/user@a.b/project/nb"
        get_this_notebook_abs_path_mock.return_value = notebook_abs_path

        ch = com.NotebooksChannel()
        ch.share_args(
            notebook_abs_path,
            {"a": "1"}
        )
        ch.share_inputs(
            notebook_abs_path,
            {"c": "3"}
        )
        ch.share_outputs(
            notebook_abs_path,
            {"e": "5"}
        )

        args = init_args("a", "b")
        inputs = init_inputs("c", "d")
        outputs = init_outputs("e", "f")

        self.assertEqual(args["a"].value, "1")
        self.assertTrue(args["a"].has_value)
        # There is no value for "b" shared.
        self.assertEqual(args["b"].value, "")
        self.assertTrue(args["b"].has_value)

        self.assertEqual(inputs["c"].value, "3")
        self.assertTrue(inputs["c"].has_value)
        # There is no value for "d" shared.
        self.assertEqual(inputs["d"].value, "")
        self.assertFalse(inputs["d"].has_value)

        self.assertEqual(outputs["e"].value, "5")
        self.assertTrue(outputs["e"].has_value)
        # There is no value for "f" shared.
        self.assertEqual(outputs["f"].value, "")
        self.assertFalse(outputs["f"].has_value)



import unittest
from unittest.mock import MagicMock, patch
from dtflw.databricks import get_notebook_abs_path, get_path_relative_to_project_dir, get_this_notebook_abs_cwd
from ddt import ddt, data, unpack


@ddt
class DatabricksTestCase(unittest.TestCase):

    @patch("dtflw.databricks.get_this_notebook_abs_path")
    def test_get_this_notebook_abs_cwd(self, get_this_notebook_abs_path_mock: MagicMock):

        get_this_notebook_abs_path_mock.return_value = "/Users/a@b.c/project_a/main"

        self.assertEqual(
            "/Users/a@b.c/project_a",
            get_this_notebook_abs_cwd())

    @patch("dtflw.databricks.get_this_notebook_abs_path")
    def test_get_notebook_abs_path(self, get_this_notebook_abs_path_mock: MagicMock):
        get_this_notebook_abs_path_mock.return_value = "/Users/a@b.c/project_b/main"

        self.assertEqual(
            "/Users/a@b.c/project_b/subdir/child",
            get_notebook_abs_path("subdir/child"))

    @data(
        ("/Users/a@b.c/main", "dir/nb", "dir/nb"),
        ("/Users/a@b.c/main", "nb", "nb"),
        ("/Users/a@b.c/project_c/main", "dir/nb", "project_c/dir/nb"),
        ("/Users/a@b.c/project_c/subproject/main", "nb", "project_c/subproject/nb")
    )
    @unpack
    @patch("dtflw.databricks.get_this_notebook_abs_path")
    def test_get_path_relative_to_project_dir(self, this_nb_abs_path, rel_path, expected_project_based_nb_path, get_this_notebook_abs_path_mock: MagicMock):
        get_this_notebook_abs_path_mock.return_value = this_nb_abs_path

        self.assertEqual(
            expected_project_based_nb_path,
            get_path_relative_to_project_dir(rel_path)
        )



import typing
from pyspark.sql.session import SparkSession
from dtflw.display import DefaultDisplay
from dtflw.storage.fs import FileStorageBase
from dtflw.logger import LoggerBase
from dtflw.pipeline import PipelineState
from dtflw.tables_repo import TablesRepository


class FlowContext():
    """
    Provides a context to all Flow subsystems to function. 
    Also, keeps and resolves evaluated at runtime tables.
    """

    def __init__(self, storage: FileStorageBase, spark: SparkSession, dbutils, logger: LoggerBase):
        self.storage = storage
        self.spark = spark
        self.dbutils = dbutils
        self.tables_repo = TablesRepository()
        self.logger = logger
        self.pipeline = PipelineState()
        self.display = DefaultDisplay()

    @property
    def logger(self) -> LoggerBase:
        """
        Returns logger object.
        """
        return self.__logger

    @logger.setter
    def logger(self, value):
        """
        Sets logger object.
        """
        if isinstance(value, LoggerBase):
            self.__logger = value
        else:
            raise ValueError("Unexpected logger type.")

    def publish_tables(self, tables: typing.Dict[str, str], src_nb_path: str) -> None:
        """
        Registers tables available to be resolved for a notebook as inputs.

        Parameters
        ----------
        tables: Dict[str, str] 
            A dictionary where key is a table's name and value is an absolute file path.

        src_nb_path: str
            Name (path) of a notebook which produced the tables.
        """
        self.tables_repo.publish(tables, src_nb_path)

    def resolve_table(self, name: str, trg_nb_path: str = None) -> str:
        """
        Searches for a table to be returned as an input for a notebook.
        Returns None if no suitable table was found.

        Parameters
        ----------
        name: str
            Name of a table to search for.
        trg_nb_path: str (None)
            Name (path) of a notebook which will consume the table as an input.

        Returns
        -------
        str
            Absolute path to the table if it has been found, and None otherwise.
        """
        return self.tables_repo.resolve(name, trg_nb_path)

    def show(self):
        """
        Show all published tables.
        """

        c = "Evaluated tables:"
        for (t_name, t_pubs) in self.tables_repo.tables.items():
            c += f"\n'{t_name}'"
            for t_nb_path, t_abs_path in t_pubs.items():
                c += f"\n  '{t_nb_path}' : '{t_abs_path}'"

        self.display.show(c)



import unittest
from unittest.mock import patch
from ddt import ddt, data
from dtflw.logger import DefaultLogger
from dtflw.flow_context import FlowContext
from dtflw.output_table import OutputTable
import tests.utils as utils


@ddt
class OutputTableTestCase(unittest.TestCase):

    @patch("pyspark.sql.session.SparkSession")
    def test_validate_succeeds(self, spark_class_mock):

        # Arrange

        dbutils_mock = utils.mock_dbutils()
        spark_mock = utils.mock_spark(spark_class_mock)

        o = OutputTable(
            name="foo",
            abs_file_path="foo.parquet",
            cols=[("id", "bigint")],
            ctx=FlowContext(
                storage=utils.StorageMock("", spark_mock, dbutils_mock),
                spark=spark_mock,
                dbutils=dbutils_mock,
                logger=DefaultLogger()
            )
        )

        # Act/Assert

        # Output is valid means that the file exists and columns are compatible.
        o.validate()

    @patch("pyspark.sql.session.SparkSession")
    def test_validate_fails_file_does_not_exist(self, spark_class_mock):

        # Arrange

        dbutils_mock = utils.mock_dbutils(fs_ls_raises_error=True)
        spark_mock = utils.mock_spark(spark_class_mock)

        o = OutputTable(
            name="foo",
            abs_file_path="foo.parquet",
            cols=[("id", "bigint")],
            ctx=FlowContext(
                storage=utils.StorageMock("", spark_mock, dbutils_mock),
                spark=spark_mock,
                dbutils=dbutils_mock,
                logger=DefaultLogger()
            )
        )

        # Act/Assert

        with self.assertRaises(Exception):
            o.validate()

    @patch("pyspark.sql.session.SparkSession")
    def test_validate_fails_columns_are_incompatible(self, spark_class_mock):

        # Arrange

        dbutils_mock = utils.mock_dbutils(fs_ls_raises_error=False)
        spark_mock = utils.mock_spark(spark_class_mock)

        o = OutputTable(
            name="foo",
            abs_file_path="foo.parquet",
            # Expected columns
            cols=[("name", "string")],
            ctx=FlowContext(
                storage=utils.StorageMock("", spark_mock, dbutils_mock),
                spark=spark_mock,
                dbutils=dbutils_mock,
                logger=DefaultLogger()
            )
        )

        # Act/Assert

        with self.assertRaises(AssertionError):
            o.validate()

    @data(True, False)
    def test_needs_eval(self, expected_needs_eval):

        # Arrange

        dbutils_mock = utils.mock_dbutils(
            fs_ls_raises_error=expected_needs_eval)

        o = OutputTable(
            name="foo",
            abs_file_path="foo.parquet",
            cols=None,
            ctx=FlowContext(
                storage=utils.StorageMock("", None, dbutils_mock),
                spark=None,
                dbutils=dbutils_mock,
                logger=DefaultLogger()
            )
        )

        self.assertEqual(o.needs_eval(), expected_needs_eval)



import dtflw.arguments as a
from dtflw.flow_context import FlowContext
from dtflw.flow import Flow
import dtflw.databricks as db
from dtflw.storage.fs import FileStorageBase
from dtflw.logger import DefaultLogger, LoggerBase
from pyspark.sql import SparkSession


def init_flow(storage: FileStorageBase, spark: SparkSession = None, dbutils=None, logger: LoggerBase = None) -> Flow:
    """
    Initializes a new, configured by default, instance of Flow and returns it.
    """

    if spark is None:
        spark = db.get_spark_session()

    if dbutils is None:
        dbutils = db.get_dbutils()

    if logger is None:
        logger = DefaultLogger()

    ctx = FlowContext(storage, spark, dbutils, logger)

    return Flow(ctx)


def init_args(*args):
    """
    Initializes args

    Parameters
    ----------
    *args
        Option 1:
            Given as a variable number of non-keyword arguments, e.g. `args = init_args("a", "b", "c")`.
            All the args will have an empty string as a value.
        Option 2:
            Given as a dictionary with default values, e.g. `args = init_args({'a': 42})`.
            All non-string values will be stringified.

    Returns
    -------
    dict[str: Argument]
    """
    return a.Argument.create(*args)


def init_inputs(*inputs):
    """
    Initializes input tables.

    Parameters
    ----------
    *inputs
        Option 1:
            Given as a variable number of non-keyword arguments, e.g. `inputs = init_inputs("Orders", "Customers")`.
            All the inputs will have an empty string as a value.
        Option 2:
            Given as a dictionary with default values, e.g. `inputs = init_inputs({'Orders': 'orders.parquet'})`.
            All non-string values will be stringified.

    Returns
    -------
    dict[str: Input]
    """
    return a.Input.create(*inputs)


def init_outputs(*outputs):
    """
    Initializes output tables.

    Parameters
    ----------
    *outputs
        Option 1:
            Given as a variable number of non-keyword arguments, e.g. `outputs = init_outputs("Orders", "Customers")`.
            All the outputs will have an empty string as a value.
        Option 2:
            Given as a dictionary with default values, e.g. `outputs = init_outputs({'Orders': 'orders.parquet'})`.
            All non-string values will be stringified.

    Returns
    -------
    dict[str: Output]
    """
    return a.Output.create(*outputs)



from abc import ABC, abstractmethod


class LoggerBase(ABC):
    """
    A base class to implement a Flow logger.
    """

    def __init__(self, verbosity: str = "default"):
        self.verbosity = verbosity

    @property
    def verbosity(self):
        return self.__verbosity

    @verbosity.setter
    def verbosity(self, value):
        """
        Posible values are
        - "default": log only errors.
        - "verbose": log errors as well as info messages.
        """
        if value in ("verbose", "default"):
            self.__verbosity = value
        else:
            self.__verbosity = "default"

    @abstractmethod
    def log(self, msg: str):
        """
        Logs an info message.
        """
        raise NotImplementedError()


class DefaultLogger(LoggerBase):
    """
    Default logger which logs to stdout using `print`.
    """

    def __log(self, msg):
        """
        Logs a message.
        """
        print(msg)

    def log(self, msg: str):
        """
        Logs an info message.
        """
        if self.verbosity == "verbose":
            self.__log(msg)



import unittest
from dtflw.display import DefaultDisplay
import sys
from io import StringIO


class DefaultDisplayTestCase(unittest.TestCase):

    def test_show(self):

        display = DefaultDisplay()

        out = StringIO()
        sys.stdout = out
        display.show('hello world!')
        output = out.getvalue().strip()

        # Assert
        self.assertEqual(output, "hello world!")
